@@ -12,11 +12,12 @@
 # language governing permissions and limitations under the License.
 """Module for processing CLI args."""
 import os
-import csv
 import json
 import logging
 import six
 
+from awscli import utils
+
 
 SCALAR_TYPES = set([
     'string', 'float', 'integer', 'long', 'boolean', 'double',
@@ -236,7 +237,7 @@ class ParamShorthand(object):
         valid_names = self._create_name_to_params(param)
         for part in parts:
             try:
-                key, value = part.split('=')
+                key, value = part.split('=', 1)
             except ValueError:
                 raise ParamSyntaxError(part)
             key = key.strip()
@@ -296,44 +297,10 @@ class ParamShorthand(object):
         return s
 
     def _split_on_commas(self, value):
-        if '"' not in value and '\\' not in value and "'" not in value:
-            # No quotes or escaping, just use a simple split.
-            return value.split(',')
-        elif '"' not in value and "'" not in value:
-            # Simple escaping, let the csv module handle it.
-            return list(csv.reader(six.StringIO(value), escapechar='\\'))[0]
-        else:
-            # If there's quotes for the values, we have to handle this
-            # ourselves.
-            return self._split_with_quotes(value)
-
-    def _split_with_quotes(self, value):
-        parts = list(csv.reader(six.StringIO(value), escapechar='\\'))[0]
-        iter_parts = iter(parts)
-        new_parts = []
-        for part in iter_parts:
-            if part.count('"') == 1:
-                quote_char = '"'
-            elif part.count("'") == 1:
-                quote_char = "'"
-            else:
-                new_parts.append(part)
-                continue
-            # Now that we've found a starting quote char, we
-            # need to combine the parts until we encounter an end quote.
-            current = part
-            chunks = [current.replace(quote_char, '')]
-            while True:
-                try:
-                    current = six.advance_iterator(iter_parts)
-                except StopIteration:
-                    raise ParamSyntaxError(value)
-                chunks.append(current.replace(quote_char, ''))
-                if quote_char in current:
-                    break
-            new_chunk = ','.join(chunks)
-            new_parts.append(new_chunk)
-        return new_parts
+        try:
+            return utils.split_on_commas(value)
+        except ValueError as e:
+            raise ParamSyntaxError(str(e))
 
 
 def unpack_cli_arg(parameter, value):
@@ -366,7 +333,8 @@ def unpack_complex_cli_arg(parameter, value):
         if value.lstrip()[0] == '{':
             d = json.loads(value)
         else:
-            msg = 'Structure option value must be JSON or path to file.'
+            msg = 'The value for parameter "%s" must be JSON or path to file.' % (
+                parameter.cli_name)
             raise ValueError(msg)
         return d
     elif parameter.type == 'list':
