@@ -2,14 +2,12 @@ import logging
 import math
 import os
 import requests
-from six import StringIO
-from six.moves import queue as Queue
-import sys
+import time
 import threading
 
-from awscli.customizations.s3.constants import QUEUE_TIMEOUT_GET
 from awscli.customizations.s3.utils import find_bucket_key, MD5Error, \
-    operate, retrieve_http_etag
+    operate, retrieve_http_etag, ReadFileChunk
+
 
 LOGGER = logging.getLogger(__name__)
 
@@ -18,12 +16,16 @@ class UploadCancelledError(Exception):
     pass
 
 
+class DownloadCancelledError(Exception):
+    pass
+
+
 def print_operation(filename, failed, dryrun=False):
     """
     Helper function used to print out what an operation did and whether
     it failed.
     """
-    print_str = filename.operation
+    print_str = filename.operation_name
     if dryrun:
         print_str = '(dryrun) ' + print_str
     if failed:
@@ -33,7 +35,7 @@ def print_operation(filename, failed, dryrun=False):
         print_str = print_str + "s3://" + filename.src
     else:
         print_str += os.path.relpath(filename.src)
-    if filename.operation not in ["delete", "make_bucket", "remove_bucket"]:
+    if filename.operation_name not in ["delete", "make_bucket", "remove_bucket"]:
         if filename.dest_type == "s3":
             print_str += " to s3://" + filename.dest
         else:
@@ -48,7 +50,7 @@ class BasicTask(object):
     attributes like ``session`` object in order for the filename to
     perform its designated operation.
     """
-    def __init__(self, session, filename, parameters, print_queue):
+    def __init__(self, session, filename, parameters, result_queue):
         self.session = session
         self.service = self.session.get_service('s3')
 
@@ -56,7 +58,7 @@ class BasicTask(object):
         self.filename.parameters = parameters
 
         self.parameters = parameters
-        self.print_queue = print_queue
+        self.result_queue = result_queue
 
     def __call__(self):
         self._execute_task(attempts=3)
@@ -70,15 +72,15 @@ class BasicTask(object):
         filename = self.filename
         try:
             if not self.parameters['dryrun']:
-                getattr(filename, filename.operation)()
+                getattr(filename, filename.operation_name)()
         except requests.ConnectionError as e:
             connect_error = str(e)
             LOGGER.debug("%s %s failure: %s",
-                         filename.src, filename.operation, connect_error)
+                         filename.src, filename.operation_name, connect_error)
             self._execute_task(attempts - 1, last_error=str(e))
         except MD5Error as e:
             LOGGER.debug("%s %s failure: Data was corrupted: %s",
-                         filename.src, filename.operation, e)
+                         filename.src, filename.operation_name, e)
             self._execute_task(attempts - 1, last_error=str(e))
         except Exception as e:
             LOGGER.debug(str(e), exc_info=True)
@@ -92,13 +94,13 @@ class BasicTask(object):
     def _queue_print_message(self, filename, failed, dryrun,
                              error_message=None):
         try:
-            if filename.operation != 'list_objects':
-                print_op = print_operation(filename, failed,
-                                           self.parameters['dryrun'])
-                print_dict = {'result': print_op}
+            if filename.operation_name != 'list_objects':
+                message = print_operation(filename, failed,
+                                          self.parameters['dryrun'])
                 if error_message is not None:
-                    print_dict['error'] = error_message
-                self.print_queue.put(print_dict)
+                    message += '\n' + error_message
+                result = {'message': message, 'error': failed}
+                self.result_queue.put(result)
         except Exception as e:
             LOGGER.debug('%s' % str(e))
 
@@ -113,8 +115,8 @@ class UploadPartTask(object):
     object.
     """
     def __init__(self, part_number, chunk_size,
-                 print_queue, upload_context, filename):
-        self._print_queue = print_queue
+                 result_queue, upload_context, filename):
+        self._result_queue = result_queue
         self._upload_context = upload_context
         self._part_number = part_number
         self._chunk_size = chunk_size
@@ -123,9 +125,8 @@ class UploadPartTask(object):
     def _read_part(self):
         actual_filename = self._filename.src
         in_file_part_number = self._part_number - 1
-        with open(actual_filename, 'rb') as in_file:
-            in_file.seek(in_file_part_number * self._chunk_size)
-            return in_file.read(self._chunk_size)
+        starting_byte = in_file_part_number * self._chunk_size
+        return ReadFileChunk(actual_filename, starting_byte, self._chunk_size)
 
     def __call__(self):
         LOGGER.debug("Uploading part %s for filename: %s",
@@ -133,39 +134,84 @@ class UploadPartTask(object):
         try:
             LOGGER.debug("Waiting for upload id.")
             upload_id = self._upload_context.wait_for_upload_id()
-            body = self._read_part()
             bucket, key = find_bucket_key(self._filename.dest)
-            if sys.version_info[:2] == (2, 6):
-                body = StringIO(body)
-            else:
-                body = bytearray(body)
+            total = int(math.ceil(
+                self._filename.size/float(self._chunk_size)))
+            body = self._read_part()
             params = {'endpoint': self._filename.endpoint,
                       'bucket': bucket, 'key': key,
                       'part_number': str(self._part_number),
                       'upload_id': upload_id,
                       'body': body}
-            response_data, http = operate(
-                self._filename.service, 'UploadPart', params)
+            try:
+                response_data, http = operate(
+                    self._filename.service, 'UploadPart', params)
+            finally:
+                body.close()
             etag = retrieve_http_etag(http)
             self._upload_context.announce_finished_part(
                 etag=etag, part_number=self._part_number)
 
-            print_str = print_operation(self._filename, 0)
-            print_result = {'result': print_str}
-            total = int(math.ceil(
-                self._filename.size/float(self._chunk_size)))
-            part_str = {'total': total}
-            print_result['part'] = part_str
-            self._print_queue.put(print_result)
+            message = print_operation(self._filename, 0)
+            result = {'message': message, 'total_parts': total,
+                      'error': False}
+            self._result_queue.put(result)
+        except UploadCancelledError as e:
+            # We don't need to do anything in this case.  The task
+            # has been cancelled, and the task that cancelled the
+            # task has already queued a message.
+            LOGGER.debug("Not uploading part, task has been cancelled.")
         except Exception as e:
             LOGGER.debug('Error during part upload: %s' , e,
                          exc_info=True)
+            message = print_operation(self._filename, failed=True,
+                                      dryrun=False)
+            message += '\n' + str(e)
+            result = {'message': message, 'error': True}
+            self._result_queue.put(result)
             self._upload_context.cancel_upload()
         else:
             LOGGER.debug("Part number %s completed for filename: %s",
                      self._part_number, self._filename.src)
 
 
+class CreateLocalFileTask(object):
+    def __init__(self, context, filename):
+        self._context = context
+        self._filename = filename
+
+    def __call__(self):
+        dirname = os.path.dirname(self._filename.dest)
+        if not os.path.isdir(dirname):
+            os.makedirs(dirname)
+        # Always create the file.  Even if it exists, we need to
+        # wipe out the existing contents.
+        with open(self._filename.dest, 'wb'):
+            pass
+        self._context.announce_file_created()
+
+
+class CompleteDownloadTask(object):
+    def __init__(self, context, filename, result_queue, params):
+        self._context = context
+        self._filename = filename
+        self._result_queue = result_queue
+        self._parameters = params
+
+    def __call__(self):
+        # When the file is downloading, we have a few things we need to do:
+        # 1) Fix up the last modified time to match s3.
+        # 2) Tell the result_queue we're done.
+        self._context.wait_for_completion()
+        last_update_tuple = self._filename.last_update.timetuple()
+        mod_timestamp = time.mktime(last_update_tuple)
+        os.utime(self._filename.dest, (int(mod_timestamp), int(mod_timestamp)))
+        message = print_operation(self._filename, False,
+                                  self._parameters['dryrun'])
+        print_task = {'message': message, 'error': False}
+        self._result_queue.put(print_task)
+
+
 class DownloadPartTask(object):
     """
     This task downloads and writes a part to a file.  This task pulls
@@ -174,74 +220,75 @@ class DownloadPartTask(object):
     in order to keep track and complete the multipart download initiated by
     the ``FileInfo`` object.
     """
-    def __init__(self, session, executer, part_queue, dest_queue,
-                 f, region, print_queue, write_lock, part_counter,
-                 counter_lock):
-        self.session = session
-        self.service = self.session.get_service('s3')
-        self.endpoint = self.service.get_endpoint(region)
-        self.executer = executer
-        self.part_queue = part_queue
-        self.dest_queue = dest_queue
-        self.f = f
-        self.print_queue = print_queue
-        self.write_lock = write_lock
-        self.part_counter = part_counter
-        self.counter_lock = counter_lock
+
+    # Amount to read from response body at a time.
+    ITERATE_CHUNK_SIZE = 1024 * 1024
+
+    def __init__(self, part_number, chunk_size, result_queue, service,
+                 filename, context):
+        self._part_number = part_number
+        self._chunk_size = chunk_size
+        self._result_queue = result_queue
+        self._filename = filename
+        self._service = filename.service
+        self._context = context
 
     def __call__(self):
+        total_file_size = self._filename.size
+        start_range = self._part_number * self._chunk_size
+        if self._part_number == int(total_file_size / self._chunk_size) - 1:
+            end_range = ''
+        else:
+            end_range = start_range + self._chunk_size - 1
+        range_param = 'bytes=%s-%s' % (start_range, end_range)
+        LOGGER.debug("Downloading bytes range of %s for file %s", range_param,
+                     self._filename.dest)
+        bucket, key = find_bucket_key(self._filename.src)
+        params = {'endpoint': self._filename.endpoint, 'bucket': bucket,
+                  'key': key, 'range': range_param}
         try:
-            part_info = self.part_queue.get(True, QUEUE_TIMEOUT_GET)
-            with self.counter_lock:
-                self.part_counter.count += 1
-            filename = part_info[0]
-            part_number = part_info[1]
-            size_uploads = part_info[2]
-            last_part_number = int(filename.size / size_uploads) - 1
-            beginning_range = part_number*size_uploads
-            str_range = "bytes="
-            if part_number == last_part_number:
-                str_range += str(beginning_range) + "-"
-            else:
-                end_range = beginning_range + size_uploads - 1
-                str_range += str(beginning_range) + "-" + str(end_range)
-            bucket, key = find_bucket_key(filename.src)
-            try:
-                params = {'endpoint': self.endpoint, 'bucket': bucket,
-                          'key': key, 'range': str_range}
-                response_data, http = operate(self.service, 'GetObject',
-                                              params)
-                body = response_data['Body'].read()
-                with self.write_lock:
-                    self.f.seek(part_number*size_uploads)
-                    self.f.write(body)
+            LOGGER.debug("Making GetObject requests with byte range: %s",
+                         range_param)
+            response_data, http = operate(self._service, 'GetObject',
+                                          params)
+            LOGGER.debug("Response received from GetObject")
+            body = response_data['Body']
+            self._write_to_file(body)
+            self._context.announce_completed_part(self._part_number)
+
+            message = print_operation(self._filename, 0)
+            total_parts = int(self._filename.size / self._chunk_size)
+            result = {'message': message, 'error': False,
+                      'total_parts': total_parts}
+            self._result_queue.put(result)
+        except Exception as e:
+            LOGGER.debug(
+                'Exception caught downloading byte range: %s',
+                e, exc_info=True)
+            self._context.cancel()
+            raise e
+
+    def _write_to_file(self, body):
+        self._context.wait_for_file_created()
+        LOGGER.debug("Writing part number %s to file: %s",
+                     self._part_number, self._filename.dest)
+        iterate_chunk_size = self.ITERATE_CHUNK_SIZE
+        with open(self._filename.dest, 'rb+') as f:
+            f.seek(self._part_number * self._chunk_size)
+            current = body.read(iterate_chunk_size)
+            while current:
+                f.write(current)
+                current = body.read(iterate_chunk_size)
+        LOGGER.debug("Done writing part number %s to file: %s",
+                     self._part_number, self._filename.dest)
 
-                print_str = print_operation(filename, 0)
-                print_result = {'result': print_str}
-                part_str = {'total': int(filename.size / size_uploads)}
-                print_result['part'] = part_str
-                self.print_queue.put(print_result)
-                self.dest_queue.put(part_number)
-            except requests.ConnectionError as e:
-                connect_error = str(e)
-                LOGGER.debug("%s part download failure: %s" %
-                            (part_info[0].src, connect_error))
-                self.part_queue.put(part_info)
-                self.executer.submit(self)
-            except Exception as e:
-                LOGGER.debug('%s' % str(e))
-            self.part_queue.task_done()
-            with self.counter_lock:
-                self.part_counter.count -= 1
-        except Queue.Empty:
-            pass
 
 
 class CreateMultipartUploadTask(BasicTask):
-    def __init__(self, session, filename, parameters, print_queue,
+    def __init__(self, session, filename, parameters, result_queue,
                  upload_context):
         super(CreateMultipartUploadTask, self).__init__(
-            session, filename, parameters, print_queue)
+            session, filename, parameters, result_queue)
         self._upload_context = upload_context
 
     def __call__(self):
@@ -255,18 +302,34 @@ class CreateMultipartUploadTask(BasicTask):
             LOGGER.debug("Error trying to create multipart upload: %s",
                          e, exc_info=True)
             self._upload_context.cancel_upload()
-            print_op = print_operation(self.filename, True,
-                                        self.parameters['dryrun'])
-            print_task = {'result': print_op, 'error': str(e)}
-            self.print_queue.put(print_task)
+            message = print_operation(self.filename, True,
+                                      self.parameters['dryrun'])
+            message += '\n' + str(e)
+            result = {'message': message, 'error': True}
+            self.result_queue.put(result)
             raise e
 
 
+class RemoveRemoteObjectTask(object):
+    def __init__(self, filename, context):
+        self._context = context
+        self._filename = filename
+
+    def __call__(self):
+        LOGGER.debug("Waiting for download to finish.")
+        self._context.wait_for_completion()
+        bucket, key = find_bucket_key(self._filename.src)
+        params = {'endpoint': self._filename.endpoint,
+                  'bucket': bucket, 'key': key}
+        response_data, http = operate(
+            self._filename.service, 'DeleteObject', params)
+
+
 class CompleteMultipartUploadTask(BasicTask):
-    def __init__(self, session, filename, parameters, print_queue,
+    def __init__(self, session, filename, parameters, result_queue,
                  upload_context):
         super(CompleteMultipartUploadTask, self).__init__(
-            session, filename, parameters, print_queue)
+            session, filename, parameters, result_queue)
         self._upload_context = upload_context
 
     def __call__(self):
@@ -287,18 +350,38 @@ class CompleteMultipartUploadTask(BasicTask):
         except Exception as e:
             LOGGER.debug("Error trying to complete multipart upload: %s",
                          e, exc_info=True)
-            print_task = {
-                'result': print_operation(self.filename, failed=True,
-                                          dryrun=self.parameters['dryrun']),
-                'error': str(e)
+            message = print_operation(
+                self.filename, failed=True,
+                dryrun=self.parameters['dryrun'])
+            message += '\n' + str(e)
+            result = {
+                'message': message,
+                'error': True
             }
         else:
             LOGGER.debug("Multipart upload completed for: %s",
                         self.filename.src)
-            print_op = print_operation(self.filename, False,
-                                       self.parameters['dryrun'])
-            print_task = {'result': print_op}
-        self.print_queue.put(print_task)
+            message = print_operation(self.filename, False,
+                                      self.parameters['dryrun'])
+            result = {'message': message, 'error': False}
+            self._upload_context.announce_completed()
+        self.result_queue.put(result)
+
+
+class RemoveFileTask(BasicTask):
+    def __init__(self, local_filename, upload_context):
+        self._local_filename = local_filename
+        self._upload_context = upload_context
+        # This 'filename' attr has to be here because other objects
+        # introspect tasks objects.  This should eventually be removed
+        # but it's needed for now.
+        self.filename = None
+
+    def __call__(self):
+        LOGGER.debug("Waiting for upload to complete.")
+        self._upload_context.wait_for_completion()
+        LOGGER.debug("Removing local file: %s", self._local_filename)
+        os.remove(self._local_filename)
 
 
 class MultipartUploadContext(object):
@@ -333,6 +416,7 @@ class MultipartUploadContext(object):
         self._lock = threading.Lock()
         self._upload_id_condition = threading.Condition(self._lock)
         self._parts_condition = threading.Condition(self._lock)
+        self._upload_complete_condition = threading.Condition(self._lock)
         self._state = self._UNSTARTED
 
     def announce_upload_id(self, upload_id):
@@ -362,6 +446,13 @@ class MultipartUploadContext(object):
                 self._upload_id_condition.wait(timeout=1)
             return self._upload_id
 
+    def wait_for_completion(self):
+        with self._upload_complete_condition:
+            while not self._state == self._COMPLETED:
+                if self._state == self._CANCELLED:
+                    raise UploadCancelledError("Upload has been cancelled.")
+                self._upload_complete_condition.wait(timeout=1)
+
     def cancel_upload(self, canceller=None, args=None, kwargs=None):
         """Cancel the upload.
 
@@ -420,5 +511,65 @@ class MultipartUploadContext(object):
         This should be called after a CompleteMultipartUpload operation.
 
         """
-        with self._lock:
+        with self._upload_complete_condition:
             self._state = self._COMPLETED
+            self._upload_complete_condition.notifyAll()
+
+
+class MultipartDownloadContext(object):
+
+    _STATES = {
+        'UNSTARTED': 'UNSTARTED',
+        'STARTED': 'STARTED',
+        'COMPLETED': 'COMPLETED',
+        'CANCELLED':'CANCELLED'
+    }
+
+    def __init__(self, num_parts, lock=None):
+        self.num_parts = num_parts
+
+        if lock is None:
+            lock = threading.Lock()
+        self._lock = lock
+        self._created_condition = threading.Condition(self._lock)
+        self._completed_condition = threading.Condition(self._lock)
+        self._state = self._STATES['UNSTARTED']
+        self._finished_parts = set()
+
+    def announce_completed_part(self, part_number):
+        with self._completed_condition:
+            self._finished_parts.add(part_number)
+            if len(self._finished_parts) == self.num_parts:
+                self._state = self._STATES['COMPLETED']
+                self._completed_condition.notifyAll()
+
+    def announce_file_created(self):
+        with self._created_condition:
+            self._state = self._STATES['STARTED']
+            self._created_condition.notifyAll()
+
+    def wait_for_file_created(self):
+        with self._created_condition:
+            while not self._state == self._STATES['STARTED']:
+                if self._state == self._STATES['CANCELLED']:
+                    raise DownloadCancelledError("Download has been cancelled.")
+                self._created_condition.wait(timeout=1)
+
+    def wait_for_completion(self):
+        with self._completed_condition:
+            while not self._state == self._STATES['COMPLETED']:
+                if self._state == self._STATES['CANCELLED']:
+                    raise DownloadCancelledError("Download has been cancelled.")
+                self._completed_condition.wait(timeout=1)
+
+    def cancel(self):
+        with self._lock:
+            self._state = self._STATES['CANCELLED']
+
+    def is_cancelled(self):
+        with self._lock:
+            return self._state == self._STATES['CANCELLED']
+
+    def is_started(self):
+        with self._lock:
+            return self._state == self._STATES['STARTED']
